type TInput
  as integer  posX, posY, _ ' left upper corner (prompt + input)
              valueX, _     ' start of the input line
              fSize, _      ' font size
              cursor, _     ' position of the cursor
              maxLength, _  ' maximum number of chars (UTF-8)
              maxSize, _    ' maximum size in pixel (prompt + input)
              reservedSpc   ' if maxSize is not set: size of prompt + 'empty string' in pixel
  as TWString prompt, _     ' prompt string
              value, _      ' input string
              empty, _      ' is shown if the input string is empty
              allowedKeys   ' string including all allowed keys for the input
  declare constructor
  declare constructor(x as integer, y as integer, s as integer, p as string, v as string, _
                      ml as integer, ms as integer = -1, e as string = "", a as string = "")
  declare sub check(key as string, eventID as integer, posSpecialKeys as integer, specialKeys as TWString)
  declare sub show(showCurs as integer = -1)
end type

declare function charLenLeft(txt as string) as integer
declare function charLenRight(txt as string) as integer
declare function checkInput(entry as TWString, maxLength as integer, byref cursor as integer = -1, _
                            key as string, eventID as integer, posSpecialKeys as integer, specialKeys as TWString, _
                            maxSize as integer = -1, fSize as integer = FONTSIZE.NORMAL, allowed as TWString = "") as string
declare sub massInput(inArr() as TInput, key as string, eventID as integer, _
                      posSpecialKeys as integer, specialKeys as TWString, _
                      byref position as integer = -1, max as integer = -1)
declare sub showCursor(x as integer, y as integer, txt as TWString, position as integer, _
                       font as FT_FaceFam, size as integer = FONTSIZE.NORMAL)

constructor TInput
end constructor
constructor TInput(x as integer, y as integer, s as integer, p as string, v as string, _
                   ml as integer, ms as integer = -1, e as string = "", a as string = "")
  posX        = x
  posY        = y
  fSize       = s
  prompt      = p
  value       = v
  cursor      = wordlength(value)
  valueX      = posX + FTsize(p, font, fSize).x
  maxLength   = ml
  maxSize     = ms
  empty       = e
  allowedKeys = a
  reservedSpc = iif(maxSize > 0, maxSize, FTsize(empty, font, fsize).x)
end constructor

sub TInput.check(key as string, eventID as integer, posSpecialKeys as integer, specialKeys as TWString)
  ' key:            pressed key
  ' eventID:        current 'event'.id (to check special keys)
  ' posSpecialKeys: position of first special key in 'event'
  ' specialKeys:    contains all special keys (as used in 'event')
  dim as integer size = maxSize+posX-valueX
  value = checkInput(value, maxLength, cursor, key, eventID, posSpecialKeys, specialKeys, size, fSize, allowedKeys)
end sub

sub TInput.show(showCurs as integer = -1)
  dim as TCoord coord = FTSize(value, font, fSize)
  if reservedSpc < coord.x + 20 then reservedSpc = coord.x + 20
  screenlock
  line (posX, posY)-step(reservedSpc, coord.y), rgb(255, 255, 255), bf
  printFT posX, posY, prompt, font, fSize
  if value = "" then
    if showCurs = 0 then printFT valueX, posY, empty, font, fSize, &h7f7f7f
  else
    printFT valueX, posY, value, font, fSize
  end if
  if showCurs then showCursor valueX, posY, value, cursor, font, fSize
  screenunlock
end sub

function charLenLeft(txt as string) as integer
  ' returns the length of the first UTF-8 char of 'txt'
  if len(txt) = 0 then return 0
  dim as integer ret = 0
  select case txt[0]
    case   0 to 127 : ret = 1
    case 193 to 208 : ret = 2
    case 224 to 232 : ret = 3
    case 240 to 244 : ret = 4
    case 248 to 250 : ret = 5
    case 252 to 253 : ret = 6
    case 254        : ret = 7
    case 255        : ret = 8
  end select
  return iif(ret > len(txt), len(txt), ret)
end function

function charLenRight(txt as string) as integer
  ' returns the length of the last UTF-8 char of 'txt'
  dim as integer position = len(txt)
  if position > 0 then
    do
      position -= 1
    loop until txt[position] < 128 or txt[position] > 191 or position = 0
  end if
  return len(txt) - position
end function

function checkInput(entry as TWString, maxLength as integer, byref cursor as integer = -1, _
                    key as string, eventID as integer, posSpecialKeys as integer, specialKeys as TWString, _
                    maxSize as integer = -1, fSize as integer = FONTSIZE.NORMAL, allowed as TWString = "") as string
  ' function for user text input
  ' returns the new entry text
  ' entry:          current entry string
  ' maxlength:      maximum number of chars for entry string
  ' cursor:         position of text cursor
  ' key:            pressed key
  ' eventID:        current 'event'.id (to check special keys)
  ' posSpecialKeys: position of first special key in 'event'
  ' specialKeys:    contains all special keys (as used in 'event')
  ' maxSize:        maximum size in pixel
  dim as TWString newEntry = entry
  dim as integer oldCursor = cursor, keyAsc = asc(key)
  if len(key) > 1 andalso keyAsc <> 255 then
    ' 'key' seems to be a UTF-8 string
    dim as TWString tempTw = key
    keyAsc = asc(*tempTw.value)
  end if
  if cursor < 0 or cursor > len(entry) then cursor = len(entry)
  if multikey(FB.SC_CONTROL) = 0 and multikey(FB.SC_ALT) = 0 then
    select case keyAsc
      case 32 to 127
        if len(allowed) andalso instr(allowed, key) = 0 then exit select
        if maxLength < 0 or len(entry) < maxlength then
          newEntry = entry.mid(1, cursor) & key & entry.mid(cursor+1)
          if maxSize < 0 orelse FTsize(newEntry, font, fSize).x < maxSize then
            cursor += 1
            return newEntry
          else
            return entry
          end if
        end if
      case 8
        if cursor > 0 then
          cursor -= 1
          return entry.mid(1, cursor) & entry.mid(cursor+2)
        end if
      case 255
        select case asc(key, 2)
          case 71, 72  ' pos1, up
            cursor = 0
          case 75      ' left
            cursor -= 1
            if cursor < 0 then cursor = 0
          case 77      ' right
            cursor += 1
            if cursor > len(entry) then cursor = len(entry)
          case 79, 80  ' end, down
            cursor = len(entry)
          case 83      ' delete
            return entry.mid(1, cursor) & entry.mid(cursor+2)
        end select
    end select
  end if
  dim as TExtendKey extKey
  if maxlength < 0 or len(entry) < maxlength then
    if eventID >= posSpecialKeys and eventID <= posSpecialKeys + len(specialKeys)-1 then
      if len(allowed) = 0 orelse instr(allowed, TWString(specialKeys[eventID-posSpecialKeys])) then
        newEntry = entry.mid(1, cursor) & specialKeys.mid(eventID-posSpecialKeys+1, 1) & entry.mid(cursor+1)
        cursor += 1
      end if
    else
      foreach(TExtendKey, extKey, globalKeyList)
        if checkCtrlAltShift(extKey.control) = 0 then continue while
        if keyAsc = extKey.ascii andalso (len(allowed) = 0 orelse instr(allowed, extKey.value)) then
          newEntry = entry.mid(1, cursor) & extKey.value & entry.mid(cursor+1)
          cursor += 1
          exit while
        end if
      nexteach
    end if
    if maxSize < 0 orelse FTsize(newEntry, font, fSize).x < maxSize then
      entry = newEntry
    else
      cursor = oldCursor
    end if
  end if
  return entry
end function

sub massInput(inArr() as TInput, key as string, eventID as integer, _
              posSpecialKeys as integer, specialKeys as TWString, _
              byref position as integer = -1, max as integer = -1)
  if max < 0 then max = ubound(inArr)
  if position < 0 then position = lbound(inArr)
  select case key
    case chr(255, 72)
      inArr(position).show(0)
      if position = lbound(inArr) then position = max else position -= 1
      inArr(position).show
    case chr(255, 80)
      inArr(position).show(0)
      if position = max then position = lbound(inArr) else position += 1
      inArr(position).show
    case chr(9)
      inArr(position).show(0)
      if multikey(FB.SC_LSHIFT) or multikey(FB.SC_RSHIFT) then
        if position = lbound(inArr) then position = max else position -= 1
      else
        if position = max then position = lbound(inArr) else position += 1
      end if
      inArr(position).show
    case ""
      ' nothing to do
    case else
      inArr(position).check key, eventID, posSpecialKeys, specialKeys
      inArr(position).show
  end select
end sub

sub showCursor(x as integer, y as integer, txt as TWString, position as integer, _
               font as FT_FaceFam, size as integer = FONTSIZE.NORMAL)
  dim as TCoord coord = FTsize(txt.mid(1, position), font, size)
  line (x+coord.x, y)-step(0, coord.y*.7)
end sub
