#include once "TWord.bi"
mdMapDeclare(long, TWord)
mdListDeclare(long)

type TListPoint
  as ulong  id, special, addition
  as single factor
  as double position
  declare constructor
  declare constructor(i as long, p as long = 0, a as long = 0)
  declare operator cast as string
  declare sub modify(lvl as integer)
end type

constructor TListPoint
end constructor

constructor TListPoint(i as long, p as long = 0, a as long = 0)
  id       = i
  position = p
  addition = a
end constructor

operator =(l1 as TListPoint, l2 as TListPoint) as integer
  return l1.id = l2.id
end operator

operator TListPoint.cast as string
  return mkl(id) & mkl(position) & mkl(addition)
end operator

sub TListPoint.modify(lvl as integer)
  ' modify this.addition and this.factor depending on word difficulty
  if factor = 0 then factor = 2.5
  factor += 0.1-(5-lvl)*(0.08+(5-lvl)*0.02)
  if factor < 1.3 then factor = 1.3
  if lvl < 3 or addition < 1 then
    addition = 1
  elseif addition < 6 then
    addition = 6
  else
    addition *= factor
  end if
  ' After each repetition assess the quality of repetition response in 0-5 grade scale:
  ' 5 - perfect response
  ' 4 - correct response after a hesitation
  ' 3 - correct response recalled with serious difficulty
  ' 2 - incorrect response; where the correct one seemed easy to recall
  ' 1 - incorrect response; the correct one remembered
  ' 0 - complete blackout.
  ' algorithm from https://www.supermemo.com/english/ol/sm2.htm
end sub

mdListDeclare(TListPoint)

type TList
  as double lastSave
  as integer size, notChecked
  as string listname
  as mdList(TListPoint) list
  declare constructor
  declare constructor(s as string)
  declare operator let(l as TList)
  declare sub add cdecl(lst as TList = TList, lvlMin as integer, lvlMax as integer, _
                        key as string = "", value as string = "")
  declare sub add cdecl(lst as TList = TList, lessions as mdList(long), key() as string, value() as string, _
                        op as Operators = Operators.and_)
  declare sub del cdecl(lvlMin as integer, lvlMax as integer, key as string = "", value as string = "")
  declare sub del cdecl(lessions as mdList(long), key() as string, value() as string, op as Operators = Operators.and_)
  declare sub remain cdecl(key as string = "", value as string = "")
  declare sub remain cdecl(lessions as mdList(long), key() as string, value() as string, _
                           op as Operators = Operators.and_)
  declare sub mix
  declare sub load(byval ln as string = "")
  declare sub save(byval ln as string = "")
  declare sub sort
  declare sub update
end type

constructor TList
end constructor

constructor TList(s as string)
  listname = s
end constructor

operator TList.let(l as TList)
  lastSave = l.lastSave
  size     = l.size
  notChecked  = l.notChecked
  listname = l.listname
  list.clear
  list.addAll(l.list)
end operator

dim shared as mdMap(long, TWord) globallist

sub TList.add cdecl(wordlist as TList = TList, lvlMin as integer, lvlMax as integer, _
                    key as string = "", value as string = "")
  ' Adds elements of wordlist.list to this.list, if the given tag corresponds.
  ' If wordlist.list is empty, words are taken from the global wordlist.
  dim entry as TListPoint, w as TWord

  if wordlist.list.isEmpty then
    dim as mdList(long) ks = globallist.keyset
    dim as long i
    foreach(long, i, ks)
      wordlist.list.add(i)
    nexteach
  end if

  foreach(TListPoint, entry, wordlist.list)
    w = globallist.get(entry.id)
    if w.lession >= lvlMin and w.lession <= lvlMax then
      select case key
        case "word"
          if w.word = value then list.add(entry)
        case "mark"
          if w.mark = cint(value) then list.add(entry)
        case "trans"
          if w.translate.contains(value) then list.add(entry)
        case "group"
          if w.groupname = value then list.add(entry)
        case "tag"
          if w.tag.contains(value) then list.add(entry)
        case ""
          list.add(entry)
        case else
          if w.map.containsKey(key) <> 0 andalso (value = "" orelse w.map.get(key) = value) then list.add(entry)
      end select
    end if
  nexteach
  size = list.size
end sub

sub TList.add cdecl(wordlist as TList = TList, lessions as mdList(long), key() as string, value() as string, _
                    op as Operators = Operators.and_)
  ' Adds elements of wordlist.list to this.list, if the given tag corresponds.
  ' If wordlist.list is empty, words are taken from the global wordlist.
  dim entry as TListPoint, w as TWord, found as boolean, v as string

  if wordlist.list.isEmpty then
    dim as mdList(long) ks = globallist.keyset
    dim as long i
    foreach(long, i, ks)
      wordlist.list.add(i)
    nexteach
  end if

  foreach(TListPoint, entry, wordlist.list)
    w = globallist.get(entry.id)
    if ubound(key) < lbound(key) then
      if lessions.contains(w.lession) then list.add(entry)
    else
      for i as integer = lbound(key) to ubound(key)
        if i < lbound(value) or i > ubound(value) then v = "" else v = value(i)
        select case key(i)
          case "word"
            found = (w.word = v)
          case "mark"
            if len(v) then
              found = (w.mark = cint(v))
            else
              found = (w.mark > 0)
            end if
          case "trans"
            found = w.translate.contains(v)
          case "group"
            found = (w.groupname = v)
          case "tag"
            found = w.tag.contains(v)
          case else
            found = (w.map.containsKey(key(i)) <> 0) andalso (v = "" orelse w.map.get(key(i)) = v)
        end select
        if op = Operators.and_ then
          if not found then exit for
        else
          if found then exit for
        end if
      next
      if op = Operators.and_ then
        if found andalso lessions.contains(w.lession) then list.add(entry)
      elseif op = Operators.or_ then
        if found orelse lessions.contains(w.lession) then list.add(entry)
      end if
    end if
  nexteach
  size = list.size
end sub

sub TList.del cdecl(lvlMin as integer, lvlMax as integer, key as string = "", value as string = "")
  ' delete elements of this.list, if the given tag corresponds
  dim entry as TListPoint, w as TWord

  foreach(TListPoint, entry, list)
    w = globallist.get(entry.id)
    if w.lession >= lvlMin and w.lession <= lvlMax then
      select case key
        case "word"
          if w.word = value then list.remove(entry)
        case "mark"
          if w.mark = cint(value) then list.remove(entry)
        case "trans"
          if w.translate.contains(value) then list.remove(entry)
        case "group"
          if w.groupname = value then list.remove(entry)
        case "tag"
          if w.tag.contains(value) then list.remove(entry)
        case ""
          list.remove(entry)
        case else
          if w.map.containsKey(key) <> 0 andalso (value = "" orelse w.map.get(key) = value) then list.remove(entry)
      end select
    end if
  nexteach
  size = list.size
end sub

sub TList.del cdecl(lessions as mdList(long), key() as string, value() as string, op as Operators = Operators.and_)
  ' delete elements of this.list, if the given tag corresponds
  dim entry as TListPoint, w as TWord, found as boolean, v as string

  foreach(TListPoint, entry, list)
    w = globallist.get(entry.id)
    if ubound(key) < lbound(key) then
      if lessions.contains(w.lession) then list.remove(entry)
    else
      found = false
      for i as integer = lbound(key) to ubound(key)
        if i < lbound(value) or i > ubound(value) then v = "" else v = value(i)
        select case key(i)
          case "word"
            found = (w.word = v)
          case "mark"
            if len(v) then
              found = (w.mark = cint(v))
            else
              found = (w.mark > 0)
            end if
          case "trans"
            found = w.translate.contains(v)
          case "group"
            found = (w.groupname = v)
          case "tag"
            found = w.tag.contains(v)
          case else
            found = (w.map.containsKey(key(i)) <> 0) andalso (v = "" orelse w.map.get(key(i)) = v)
        end select
        if op = Operators.and_ then
          if not found then exit for
        else
          if found then list.remove(entry) : exit for
        end if
      next
      if op = Operators.and_ then
        if found andalso lessions.contains(w.lession) then list.remove(entry)
      elseif op = Operators.or_ then
        if lessions.contains(w.lession) then list.remove(entry)
      end if
    end if
  nexteach
  size = list.size
end sub

sub TList.remain cdecl(key as string = "", value as string = "")
  ' delete all elements of this.list that do not have corresponding tags
  dim entry as TListPoint, w as TWord

  foreach(TListPoint, entry, list)
    w = globallist.get(entry.id)
    select case key
      case "word"
        if w.word <> value then list.remove(entry)
      case "mark"
        if w.mark <> cint(value) then list.remove(entry)
      case "trans"
        if not w.translate.contains(value) then list.remove(entry)
      case "group"
        if w.groupname <> value then list.remove(entry)
      case "tag"
        if not w.tag.contains(value) then list.remove(entry)
      case ""
      case else
        if w.map.containsKey(key) = 0 orelse (value <> "" andalso w.map.get(key) <> value) then list.remove(entry)
    end select
  nexteach
  size = list.size
end sub

sub TList.remain cdecl(lessions as mdList(long), key() as string, value() as string, op as Operators = Operators.and_)
  ' delete all elements of this.list that do not have corresponding tags
  dim entry as TListPoint, w as TWord, found as boolean, v as string

  foreach(TListPoint, entry, list)
    w = globallist.get(entry.id)
    if not lessions.contains(w.lession) then
      list.remove(entry)
    else
      found = false
      for i as integer = lbound(key) to ubound(key)
        if i < lbound(value) or i > ubound(value) then v = "" else v = value(i)
        select case key(i)
          case "word"
            found = (w.word <> v)
          case "mark"
            if len(v) then
              found = (w.mark <> cint(v))
            else
              found = (w.mark = 0)
            end if
          case "trans"
            found = not w.translate.contains(v)
          case "group"
            found = (w.groupname <> v)
          case "tag"
            found = not w.tag.contains(v)
          case else
            found = (w.map.containsKey(key(i)) = 0) orelse (v <> "" andalso w.map.get(key(i)) <> v)
        end select
        if op = Operators.and_ then
          if not found then exit for
        else
          if found then list.remove(entry) : exit for
        end if
      next
    end if
  nexteach
  size = list.size
end sub

sub TList.mix
  ' mix elements in the list (regardless of position and addition)
  dim listPoint as TListPoint, index as integer, i as integer = 0
  foreach(TListPoint, listPoint, list)
    if listPoint.position > 0 then exit while
    i += 1
  nexteach
  for k as integer = i-1 to 0 step -1
    index = int(rnd*k)
    listPoint = list.get(index)
    list.remove(index)
    if k = size-1 then list.add(listPoint) else list.add(k, listPoint)
  next
end sub

sub TList.load(byval ln as string = "")
  list.clear
  dim as integer filenumber = freefile
  dim as TListPoint listPoint
  if ln = "" then ln = listname
  open MODULE_PATH & ln & ".lst" for binary as #filenumber
  get #filenumber,, lastSave
  do until eof(filenumber)
    get #filenumber,, listPoint
    if globallist.containsKey(listPoint.id) then list.add(listPoint)
  loop
  close #filenumber
  update
end sub

sub TList.save(byval ln as string = "")
  dim as TListPoint listPoint
  dim as integer filenumber = freefile
  if ln = "" then ln = listname
  kill MODULE_PATH & ln & ".lst"
  open MODULE_PATH & ln & ".lst" for binary as #filenumber
  put #filenumber,, lastSave
  foreach(TListPoint, listPoint, list)
    put #filenumber,, listPoint
  nexteach
  close #filenumber
end sub

sub TList.sort
  ' sort by position (if something got wrong)
  dim as TListPoint cur, last = list.get(0)
  dim as integer i, index = 1
  do while index < size
    cur = list.get(index)
    if cur.position < last.position then
      for i = index-2 to 0 step -1
        if list.get(i).position <= cur.position then exit for
      next
      list.remove(index)
      list.add(i+1, cur)
      last = cur
      index = i+2
    else
      last = cur
      index += 1
    end if
  loop
end sub

sub TList.update
  dim as TListPoint listPoint
  size = list.size
  notChecked = 0
  foreach(TListPoint, listPoint, list)
    if listPoint.position <= now then notChecked += 1 else exit while
  nexteach
end sub

