#include once "TWString.bi"
#include once "fbgfx.bi"
#include once "utf_conv.bi"
#include once "unicode2UpperLower.bi"

namespace FONTSIZE
  const NORMAL = 18, HEADLINE = 20, SMALL = 16, X_SMALL = 14, TINY = 12, BUTTON = 13
end namespace

namespace FONTFACE
  const NORMAL = 0, BOLD = 1, ITAlIC = 2, SMALLCAPS = 4 ', SUPER = 8, SUB_ = 16
  const SMALLCAPS_FACTOR = .8
end namespace

#ifdef FBGFX_FALLBACK
  ' missing symbols from freetype.bi
  function FT_Init_FreeType(p as any ptr) as integer
    return 0
  end function
  type FT_Face as integer
  /'
  end type '/
  type TFT_Var
    ErrorMsg  as integer
    Library   as integer
    PixelSize as integer
  end type
#else
  #include once "freetype2/freetype.bi"
  ' alpha blending
  const FT_MASK_RB_32 = &h00ff00ff, FT_MASK_AG_32 = &hff00ff00
  type TFT_Var
    ErrorMsg  as FT_Error
    Library   as FT_Library
    PixelSize as integer
  end type
#endif

type FT_FaceFam
  as FT_Face face(3)
  declare sub get(number as integer, filename as string, FT_Var as TFT_Var)
end type

sub FT_FaceFam.get(number as integer, fontName as string, FT_Var as TFT_Var)
  #ifndef FBGFX_FALLBACK
    FT_Var.ErrorMsg = FT_New_Face(FT_Var.Library, fontName, 0, @face(number))
    if FT_Var.ErrorMsg then face(number) = 0
  #endif
end sub

type TCoord
  as integer x, y
  declare constructor
  declare constructor(x_ as integer, y_ as integer)
end type
constructor TCoord
end constructor
constructor TCoord(x_ as integer, y_ as integer)
  x = x_
  y = y_
end constructor

declare sub drawGlyph(img as FB.Image ptr, byval font as FT_Face, byval x as integer, byval y as integer, byval clr as uinteger)
declare function FTsize(byval txt as TWString, byval font as FT_FaceFam, byval size as integer = FONTSIZE.NORMAL) as TCoord
declare function fontsLoadFromFile(filename as string, byref font as FT_FaceFam) as FT_ERROR
declare sub centerFT overload(img as FB.Image ptr, txt as TWString, font as FT_FaceFam, _
                     size as integer = FONTSIZE.NORMAL, clr as uinteger = 0, stress as integer = 0)
declare sub centerFT(x as integer, y as integer, txt as TWString, font as FT_FaceFam, _
                     size as integer = FONTSIZE.NORMAL, clr as uinteger = 0, stress as integer = 0)
declare function printFT overload(txt as TWString, font as FT_FaceFam, _
                     size as integer = FONTSIZE.NORMAL, clr as uinteger = 0, stress as boolean = false) as FB.Image ptr
declare function printFT(x as integer, y as integer, txt as TWString, font as FT_FaceFam, _
                     size as integer = FONTSIZE.NORMAL, clr as uinteger = 0, _
                     bmpPtr as FB.Image ptr = 0, stress as boolean = false) as boolean

sub drawGlyph(img as FB.Image ptr, byval font As FT_Face, byval x as integer, byval y as integer, byval clr as uinteger)
  #ifndef FBGFX_FALLBACK
    if img = 0 then exit sub
    dim as FT_Bitmap BmpFT
    dim as ubyte ptr BmpPtr
    dim as ulong ptr dstPtr, startPtr
    dim as integer   BmpHeight, BmpWidth, BmpPitch
    dim as uinteger  srcRB, srcAG, dstRB, dstAG, dstClr, alph
  
    bmpFT     = font->Glyph->Bitmap
    bmpPtr    = bmpFT.Buffer
    bmpWidth  = bmpFT.Width
    bmpHeight = bmpFT.Rows

    startPtr = cast(ulong ptr, img) + 8
    dstPtr = startPtr + y*img->Pitch\img->bpp + x
    do while bmpHeight
      do while bmpWidth
        if dstPtr >= startPtr then
          srcRB    = clr and FT_MASK_RB_32
          srcAG    = clr and FT_MASK_AG_32
          dstClr   = *dstPtr
          alph     = *bmpPtr
          dstRB    = dstClr and FT_MASK_RB_32
          dstAG    = dstClr and FT_MASK_AG_32
          srcRB    = ((srcRB - dstRB) * alph) shr 8
          srcAG    = ((srcAG  - dstAG) shr 8)  * alph
          *dstPtr  = ((dstRB + srcRB) and FT_MASK_RB_32) or ((dstAG + srcAG) and FT_MASK_AG_32)
        end if
        dstPtr   += 1
        bmpPtr   += 1
        bmpWidth -= 1
      loop
      bmpWidth  = bmpFT.Width
      bmpHeight -= 1
      dstPtr += img->pitch\img->bpp - bmpFT.Width
    loop
  #endif
end sub

function FTsize(byval txt as TWString, byval font as FT_FaceFam, byval size as integer = FONTSIZE.NORMAL) as TCoord
  #ifdef FBGFX_FALLBACK
    return type<TCoord>(len(*txtptr)*8, 16)
  #else
    dim as integer curFace = FONTFACE.NORMAL, curSize = size, newFace = curFace, newSize = curSize
    if font.face(curFace) = 0 then return type(-1, -1)
    dim as FT_Error ErrorMsg
    dim as FT_UInt GlyphIndex
    dim as boolean smallCaps = false
    dim as integer px = 2, py = 0, i = 0, tempInt
    dim as FT_GlyphSlot slot = font.face(curFace)->Glyph
    'FT_Var.PixelSize = curSize
    if FT_Set_Pixel_Sizes(font.face(curFace), curSize, curSize) then return TCoord(-1, -1)
    slot = font.face(curFace)->Glyph
    GlyphIndex = FT_Get_Char_Index(font.face(curFace), 124)
    if FT_Load_Glyph(font.face(FONTFACE.NORMAL), GlyphIndex, FT_LOAD_DEFAULT) then return TCoord(-1, -1)
    if FT_Render_Glyph(slot, FT_RENDER_MODE_NORMAL) then return type(-1, -1)
    py = slot->metrics.height\64+2*(size-slot->Bitmap_Top)

    do while i < len(txt)
      if txt[i] = 27 then   ' ESCAPE: change font style etc.
        ' \27n, \27N       : set to normal
        ' \27B, \27I, \27c : set to bold / italic / small caps
        ' \27b, \27i, \27c : switch bold / italic / small caps (on/off)
        i += 1
        if i >= len(txt) then exit do
        select case txt[i]
          case 78, 110                             ' N, n
            newFace = FONTFACE.NORMAL
            smallCaps = false
          case  66 : newFace  or= FONTFACE.BOLD    ' B
          case  67 : smallCaps  = true             ' C
          case  73 : newFace  or= FONTFACE.ITALIC  ' I
          case  98 : newFace xor= FONTFACE.BOLD    ' b
          case  99 : smallCaps  = not(smallCaps)   ' c
          case 105 : newFace xor= FONTFACE.ITALIC  ' i
        end select
        i += 1
        continue do
      end if
      #ifdef FONT_STRESS_ESCAPE
        if txt[i] = FONT_STRESS_ESCAPE then i += 1 : continue do
      #endif
      if smallCaps then
        newSize = iif(txt[i] = unicode2upper(txt[i]), size, size*FONTFACE.SMALLCAPS_FACTOR)
        GlyphIndex = FT_Get_Char_Index(font.face(newFace), unicode2upper(txt[i]))
      else
        newSize = size
        GlyphIndex = FT_Get_Char_Index(font.face(newFace), txt[i])
      end if
      if font.face(newFace) = 0 then newFace = FONTFACE.NORMAL  ' fallback
      if curSize <> newSize or curFace <> newFace then
        curSize = newSize
        curFace = newFace
        if FT_Set_Pixel_Sizes(font.face(curFace), curSize, curSize) then return TCoord(-1, -1)
        slot = font.face(curFace)->Glyph
      end if
      if FT_Load_Glyph(font.face(curFace), GlyphIndex, FT_LOAD_DEFAULT) then return TCoord(-1, -1)
      px += slot->Advance.x shr 6
      i += 1
    loop
    return type(px, py)
  #endif
end function

function fontsLoadFromFile(filename as string, byref font as FT_FaceFam) as FT_ERROR
  dim as TFT_Var FT_Var
  FT_Var.ErrorMsg = FT_Init_FreeType(@FT_Var.Library)
  if FT_Var.ErrorMsg then return FT_Var.ErrorMsg
  dim as integer search, search2, filenumber = freefile
  dim as string tempString, key, path
  open filename for input as #filenumber
  do until eof(filenumber)
    line input #filenumber, tempString
    search = instr(tempString, any chr(32, 9))
    if search then
      key  = left(tempString, search-1)
      path = trim(mid(tempString, search+1), any chr(32, 9))
      search = instr(path, "%")
      if search then
        search2 = instr(search+1, path, "%")
        if search2 then path = left(path, search-1) & environ(mid(path, search+1, search2-search-1)) & mid(path, search2+1)
      end if
      select case key
        case "NORMAL"
          font.get(FONTFACE.NORMAL, path, FT_Var)
        case "BOLD"
          font.get(FONTFACE.BOLD,   path, FT_Var)
        case "ITALIC"
          font.get(FONTFACE.ITALIC, path, FT_Var)
        case "BOLDITALIC"
          font.get(FONTFACE.BOLD or FONTFACE.ITALIC, path, FT_Var)
      end select
     end if
  loop
  close #filenumber
  return FT_Var.ErrorMsg
end function

sub centerFT(img as FB.Image ptr, txt as TWString, font as FT_FaceFam, size as integer = FONTSIZE.NORMAL, _
             clr as uinteger = 0, stress as integer = 0)
  if img = 0 then exit sub
  dim as FB.Image ptr tempImg = printFT(txt, font, size, clr, stress)
  dim as integer sx, sy, dx, dy, w, h
  if tempImg then
    if tempImg->width >= img->width then
      sx = 0
      dx = (tempImg->width-img->width)\2
      w  = img->width-1
    else
      sx = (img->width-tempImg->width)\2 + 1
      dx = 0
      w  = tempImg->width-1
    end if
    if tempImg->height-4 >= img->height then
      sy = 4
      dy = (tempImg->height-img->height)\2
      h  = img->height-1
    else
      sy = (img->height-tempImg->height+4)\2
      dy = 0
      h  = tempImg->height-1
    end if
    #ifdef FBGFX_FALLBACK
      put img, (sx, sy), tempImg, (dx, dy)-step(w, h)
    #else
      put img, (sx, sy), tempImg, (dx, dy)-step(w, h), alpha
    #endif
    imagedestroy tempImg
  end if
end sub

sub centerFT(x as integer, y as integer, txt as TWString, font as FT_FaceFam, size as integer = FONTSIZE.NORMAL, _
             clr as uinteger = 0, stress as integer = 0)
  dim as TCoord textsize = FTsize(txt, font, size)
  printFT x-textsize.x\2, y-textsize.y\2, txt, font, size, clr,, stress
end sub

function printFT(txt as TWString, font as FT_FaceFam, size as integer = FONTSIZE.NORMAL, _
                 clr as uinteger = 0, stress as boolean = false) as FB.Image ptr
  #ifdef FBGFX_FALLBACK
    dim as FB.Image ptr img = imagecreate(len(txt)*8, 12, 0)
    draw string img, (0, 0), utf2ansi(txt), not clr
  #else
    dim as integer curFace = FONTFACE.NORMAL, curSize = size, newFace = curFace, newSize = curSize
    if font.face(curFace) = 0 then return 0
    dim as TCoord imgSize = FTsize(txt, font, size)
    imgSize.x += 4
    dim as FB.Image ptr img = imagecreate(imgSize.x, imgSize.y, 0)
    dim as FT_UInt GlyphIndex
    dim as FT_GlyphSlot slot
    dim as boolean smallCaps = false
    dim as integer px = 3, py = 0
    clr or= &hff000000
    'FT_Var.PixelSize = curSize
    if FT_Set_Pixel_Sizes(font.face(curFace), curSize, curSize) then imagedestroy img : return 0
    slot = font.face(curFace)->Glyph
    #ifdef FONT_STRESS_ESCAPE
      dim as uinteger clrStress
      #ifdef FONT_STRESS_COLOR
        clrStress = FONT_STRESS_COLOR
      #else
        clrStress = clr
      #endif
    #endif

    dim as integer i = 0, curColor = clr
    do while i < len(txt)
      if txt[i] = 27 then   ' ESCAPE: change font style etc.
        ' \27n, \27N       : set to normal
        ' \27B, \27I, \27C : set to bold / italic / small caps
        ' \27b, \27i, \27c : switch bold / italic / small caps (on/off)
        i += 1
        if i >= len(txt) then exit do
        select case txt[i]
          case 78, 110                             ' N, n
            newFace = FONTFACE.NORMAL
            smallCaps = false
          case  66 : newFace  or= FONTFACE.BOLD    ' B
          case  67 : smallCaps  = true             ' C
          case  73 : newFace  or= FONTFACE.ITALIC  ' I
          case  98 : newFace xor= FONTFACE.BOLD    ' b
          case  99 : smallCaps  = not(smallCaps)   ' c
          case 105 : newFace xor= FONTFACE.ITALIC  ' i
        end select
        i += 1
        continue do
      end if
      #ifdef FONT_STRESS_ESCAPE
        if txt[i] = FONT_STRESS_ESCAPE then
          if stress then
            #ifdef FONT_STRESS_CHAR
              GlyphIndex = FT_Get_Char_Index(font.face(curFace), FONT_STRESS_CHAR)
              if FT_Load_Glyph(font.face(curFace), GlyphIndex, FT_LOAD_DEFAULT) then imagedestroy img : return 0
              if FT_Render_Glyph(slot, FT_RENDER_MODE_NORMAL) then imagedestroy img : return 0
              drawGlyph img, font.face(curFace), px + slot->Bitmap_Left, py - slot->Bitmap_Top + size, clrStress
            #endif
            curColor = clrStress
          end if
          i += 1
          if i >= len(txt) then exit do
        else
          curColor = clr
        end if
      #endif
      if smallCaps then
        newSize = iif(txt[i] = unicode2upper(txt[i]), size, size*FONTFACE.SMALLCAPS_FACTOR)
        GlyphIndex = FT_Get_Char_Index(font.face(newFace), unicode2upper(txt[i]))
      else
        newSize = size
        GlyphIndex = FT_Get_Char_Index(font.face(newFace), txt[i])
      end if
      if font.face(newFace) = 0 then newFace = FONTFACE.NORMAL  ' fallback
      if curSize <> newSize or curFace <> newFace then
        curSize = newSize
        curFace = newFace
        'FT_Var.PixelSize = curSize
        if FT_Set_Pixel_Sizes(font.face(curFace), curSize, curSize) then imagedestroy img : return 0
        slot = font.face(curFace)->Glyph
      end if
      if FT_Load_Glyph(font.face(curFace), GlyphIndex, FT_LOAD_DEFAULT) then imagedestroy img : return 0
      if FT_Render_Glyph(slot, FT_RENDER_MODE_NORMAL) then imagedestroy img : return 0
      ' set pixels
      drawGlyph img, font.face(curFace), px + slot->Bitmap_Left, py - slot->Bitmap_Top + size, curColor
      px += slot->Advance.x shr 6
      i += 1
    loop
  #endif
  return img
end function

function printFT(x as integer, y as integer, txt as TWString, font as FT_FaceFam, size as integer = FONTSIZE.NORMAL, _
                 clr as uinteger = 0, bmpPtr as FB.Image ptr = 0, stress as boolean = false) as boolean
  #ifdef FBGFX_FALLBACK
    draw string bmpPtr, (x, y), utf2ansi(txt), clr
  #else
    dim as FB.Image ptr img = printFT(txt, font, size, clr, stress)
    if img = 0 then return false
    put bmpPtr, (x, y), img, alpha
    imagedestroy img
  #endif
  return true
end function

function wordLength(w as string) as integer
  ' returns the size of a utf8-string 'w' in WString chars (e.g. wordLength("¿Qué?") returns 5)
  ' This includes stress escape characters!
  dim as TWString tempWStr = w
  return len(tempWStr)
end function
