' global functions and defines used in all modules

' #define FBGFX_FALLBACK    ' only if you cannot use freetype - the fallback option is NOT COMPLETELY SUPPORTED

enum RETVAR
  aborted = -99   ' exercise was aborted before it was finished
  windowClosed    ' exit program by the window close button
  generalError    ' general error; not specified
  missingLib      ' cannot load LIB (e. g. freetype)
  missingFont     ' cannot load font (e. g. missing or corrupt font.ttf)
  emptyList       ' wordlist is empty
  dataError       ' general data error (e. g. missing image file)
  ok = 0
end enum
enum Operators
  none = 0, and_, or_
end enum

dim shared as integer GlobalSettings
namespace GLOBAL_SETTINGS
  const ALLOW_EDITS = 1
end namespace

#include once "shared.bi"
#include once "fbthread.bi"
#include once "vbcompat.bi"
#include once "include/TList.bi"
#include once "include/TWString.bi"

dim shared as any ptr speakLock

declare sub drawProgress overload(x as integer, y as integer, w as integer, h as integer, _
                                  i as integer, all as integer)
declare sub drawProgress(x as integer, y as integer, w as integer, h as integer, _
                         progress() as integer, col() as uinteger)
declare sub flipCard(x as integer, y as integer, img1 as FB.Image ptr, img2 as FB.Image ptr = 0, _
                   bg as uinteger = &hffffff)
declare function loadDate(listname as string) as double
declare function localize(key as string, t1 as TWString = "", t2 as TWString = "", t3 as TWString = "") as TWString
declare function message(msg as TWString, byval buttons as TWString = "", byval focus as integer = 0, _
                         stress as integer = -1) as integer
declare sub messageButton(buttonCol as uinteger, txtCol as uinteger = rgb(0, 0, 0), _
                         txt as TWString = "", txtSize as integer = -1, _
                         x as integer = -1, y as integer = -1, w as integer = -1, h as integer = -1, _
                         stress as integer = -1)
declare sub messageWait(msg as TWString = "", stress as integer = -1)
declare function replace(txt as TWString, srch as TWString, repl as TWString) as TWString
declare function showHelpScreen(filename as string, chapter as string) as integer
declare sub sortArray(keyArray() as string, lpArray() as TListPoint, direction as integer = 1, _
                      byval s as integer = -1, byval e as integer = -1)
declare function sortLetters(txt as TWString) as TWString

#include once "include/multiput.bi"
#include once "include/TExtendKey.bi"
#include once "include/TExercise.bi"
#include once "include/TGui.bi"
#include once "include/help.bi"

' initiate language file
dim shared as mdMap(string, TWString) langData
scope
  dim as integer search, fileNumber = freefile
  dim as string fullLine, key, value
  open MODULE_PATH & "localize.txt" for input as filenumber
  do until eof(fileNumber)
    line input #fileNumber, fullLine
    fullLine = trim(fullLine, any chr(9, 32))
    if left(fullLine, 1) <> "#" then
      search = instr(fullLine, "=")
      if search = 0 then continue do
      langData.put(rtrim(left(fullLine, search-1), any chr(9, 32)), ltrim(mid(fullLine, search+1), any chr(9, 32)))
    end if
  loop
  close filenumber
end scope

sub drawProgress(x as integer, y as integer, w as integer, h as integer, i as integer, all as integer)
  ' draw progressing bar
  ' x, y, w, h: geometry of the progressing bar
  ' the fraction (i / all) represents how much of the work is done
  if i < 0 or i > all then exit sub
  line (x, y)-step(w, h), 0, b
  line (x+1, y+1)-step(w-2, h-2), &hffffff, bf
  line (x+1, y+1)-step(w*i/all-2, h-2), COLOR_GRAY, bf
end sub

sub drawProgress(x as integer, y as integer, w as integer, h as integer, _
                 progress() as integer, col() as uinteger)
  ' draw progressing bar
  ' x, y, w, h: geometry of the progressing bar
  ' progress(i) contains which part of the bar should be drawn in color col(i)
  dim sum as integer = 0, posX as double = x+1, curCol as uinteger
  for i as integer = lbound(progress) to ubound(progress)
    sum += progress(i)
  next
  for i as integer = lbound(progress) to ubound(progress)
    curCol = iif(lbound(col) > i or ubound(col) < i, rgb(255, 255, 255), col(i))
    line (posX, y+1)-step((w-2)*progress(i)/sum, h-2), curCol, bf
    posX += (w-2)*progress(i)/sum
  next
  line (x, y)-step(w, h), 0, b
end sub

sub flipCard(x as integer, y as integer, img1 as FB.Image ptr, img2 as FB.Image ptr = 0, bg as uinteger = &hffffff)
  ' "fold" img1 and then "fold apart" img2
  if img1 = 0 then exit sub
    for i as integer = 10 to 1 step -1
      screenlock
      line(x, y)-step(img1->width-1, img1->height-1), bg, bf
      #ifdef __FB_64BIT__
        dim as integer w = img1->width*i/20
        put (x+img1->width\2-w, y), img1, (0, 0)-step(w, img1->height), pset
        put (x+img1->width\2, y), img1, (img1->width-w, 0)-(img1->width, img1->height), pset
      #else
        multiPut , x+img1->width\2, y+img1->height\2, img1, i/10
      #endif
      screenunlock
      sleep 14
    next
  line(x, y)-step(img1->width-1, img1->height-1), bg, bf
  if img2 = 0 then exit sub
  for i as integer = 1 to 10
    screenlock
    line(x, y)-step(img2->width-1, img2->height-1), bg, bf
    #ifdef __FB_64BIT__
      dim as integer w = img2->width*i/20
      put (x+img2->width\2-w, y), img2, (0, 0)-step(w, img2->height), pset
      put (x+img2->width\2, y), img2, (img2->width-w, 0)-(img2->width, img2->height), pset
    #else
      multiPut , x+img2->width\2, y+img2->height\2, img2, i/10
    #endif
    screenunlock
    sleep 14
  next
end sub

function loadDate(listname as string) as double
  ' extract date of last edit of an exercise file
  ' returns the serial number of the date
  dim as integer filenumber = freefile
  dim as double ret
  open "/" & language & "/" & listname & ".lst" for binary as #filenumber
  get #filenumber,, ret
  close #filenumber
  return ret
end function

function localize(key as string, t1 as TWString = "", t2 as TWString = "", t3 as TWString = "") as TWString
  dim as string ret = langData.get(key)
  if len(ret) = 0 then return "<" & key & ">"
  dim as integer search
  search = instr(ret, "<%1%>")
  if search then ret = left(ret, search-1) & str(t1) & mid(ret, search+5)
  search = instr(ret, "<%2%>")
  if search then ret = left(ret, search-1) & str(t2) & mid(ret, search+5)
  search = instr(ret, "<%3%>")
  if search then ret = left(ret, search-1) & str(t3) & mid(ret, search+5)
  return ret
end function

function message(msg as TWString, byval buttons as TWString = "", byval focus as integer = 0, _
                 stress as integer = -1) as integer
  ' shows a message box with interaction
  ' focus: number of the button with focus (confirm with Enter, change with Tab)
  ' focus = 0: no button has or can get the focus
  ' return value: number of the activated button, starts with 1
  dim as integer countLines = 1, countButtons = 1, srch, mx, my, mb
  dim as integer maxLines = 20, maxButtons = 10, fSize = FONTSIZE.X_SMALL
  dim as integer msgHeight, msgWidth = 100, bWidth = 50    ' minimal width
  dim as integer maxWidth = SCREEN_X - 100
  dim as string lines(maxLines-1), butt(maxButtons-1), tempString, key
  dim as TGuiElement guiButton(1 to maxButtons)
  dim as TCoord coord
  dim as TGuiEvent event
  if buttons = "" then buttons = localize("ok")
  ' get message lines
  tempString = msg
  srch = instr(tempString, "|")
  do while srch
    lines(countLines-1) = left(tempString, srch-1)
    tempString = mid(tempString, srch+1)
    countLines += 1
    if countLines = maxLines then exit do
    srch = instr(tempString, "|")
  loop
  lines(countLines-1) = tempString
  ' get buttons
  tempString = buttons
  srch = instr(tempString, "|")
  do while srch
    butt(countButtons-1) = left(tempString, srch-1)
    tempString = mid(tempString, srch+1)
    countButtons += 1
    if countButtons = maxButtons then exit do
    srch = instr(tempString, "|")
  loop
  butt(countButtons-1) = tempString
  ' calculate geometry
  for i as integer = 0 to countLines-1
    coord = FTsize(lines(i), font, fSize)
    if coord.x > msgWidth then msgWidth = coord.x
  next
  for i as integer = 0 to countButtons-1
    coord = FTsize(butt(i), font, fSize)
    if coord.x > bWidth then bWidth = coord.x
  next
  msgHeight = coord.y*(countLines+2)
  ' extend box for buttons (if necessary)
  if msgWidth < (bWidth+30)*countButtons then msgWidth = (bWidth+30)*countButtons
  ' reduce box size if too big
  if msgWidth > maxWidth then msgWidth = maxWidth
  ' reduce buttons size if too big
  if msgWidth < (bWidth+30)*countButtons then bWidth = msgWidth\countButtons - 30
  ' draw box
  dim as FB.Image ptr lineImg, boxImg = imagecreate(msgWidth+40, msgHeight+40)
  dim as integer startX = (SCREEN_X-msgWidth)\2, startY = (SCREEN_Y-msgHeight)\2
  get (startX-20, startY-20)-step(msgWidth+39, msgHeight+39), boxImg
  line (startX-20, startY-20)-step(msgWidth+39, msgHeight+39), &hffffff, bf
  line (startX-19, startY-19)-step(msgWidth+37, msgHeight+37), 0, b
  line (startX-18, startY-18)-step(msgWidth+35, msgHeight+35), 0, b
  line (startX-16, startY-16)-step(msgWidth+31, msgHeight+31), 0, b
  ' show lines
  for i as integer = 0 to countLines-1
    lineImg = printFT(lines(i), font, fSize,, stress)
    put (startX, startY+coord.y*i), lineImg, (0, 0)-step(msgWidth, coord.y), alpha
    imagedestroy lineImg
  next
  ' show buttons
  dim as TGui messageGui
  dim as integer bStartX = (SCREEN_X - (bWidth+30)*countButtons)\2 + 10
  for i as integer = 1 to countButtons
    guiButton(i) = TGuiElement(GuiElementType.button, i, bStartX+(i-1)*(bWidth+30)-3, startY+msgHeight-10-coord.y, _
                               bWidth+15, coord.y+5, butt(i-1))
    if focus = i then guiButton(i).state or= GuiState.focus
    messageGui.addElement guiButton(i)
  next
  messageGui.show
  ' wait for user input
  do
    key = inkey
    select case key
      case chr(9)
        if focus > 0 and focus <= countButtons then
          guiButton(focus).state and= not GuiState.focus
          messageGui.setElement guiButton(focus)
          focus += 1
          if focus > countButtons then focus = 1
          guiButton(focus).state or= GuiState.focus
          messageGui.setElement guiButton(focus)
          messageGui.show
        end if
      case chr(13)
        if focus > 0 and focus <= countButtons then event.id = focus : exit do
    end select
    event = messageGui.getEvent
    if event.mButton > 0 and event.id > 0 then exit do
    sleep 1
  loop until key = chr(255, 107)
  ' clean up
  put (startX-20, startY-20), boxImg, pset
  imagedestroy boxImg
  if key = chr(255, 107) then return RETVAR.windowClosed
  do
    getmouse mx, my,, mb
    sleep 1
  loop until mb < 1
  return event.id
end function

sub messageButton(buttonCol as uinteger, txtCol as uinteger = rgb(0, 0, 0), _                   ' color
                  txt as TWString = "", txtSize as integer = -1, _                              ' text
                  x as integer = -1, y as integer = -1, w as integer = -1, h as integer = -1, _ ' geometry
                  stress as integer = -1)
  ' shows a button with text
  ' You can leave out x, y, w and h to use the values of the last call.
  ' For example, messagebutton(-1) overwrites the last messageButton with a white box.
  static as integer bx, by, bw, bh, tSize = 1
  if x >= 0 then bx = x
  if y >= 0 then by = y
  if w >= 0 then bw = w
  if h >= 0 then bh = h
  if txtSize >= 0 then tSize = txtSize
  dim as FB.Image ptr bImg = imagecreate(bw, bh, buttonCol)
  centerFT bImg, txt, font, tSize, txtCol, stress
  put (bx, by), bImg, pset
  imagedestroy bImg
end sub

sub messageWait(msg as TWString = "", stress as integer = -1)
  ' shows a message window with no interaction (such like "please wait ...")
  ' msg = "": hide message (restore background)

  ' IF YOU SHOW A MESSAGE, YOU HAVE TO FREE MEMORY BY AN EMPTY MESSAGE!
  ' A new message deletes an old one; no need to call an empty message between.
  static as FB.Image ptr boxImg
  static as integer startX, startY
  if boxImg then
    put (startX-20, startY-20), boxImg, pset
    imagedestroy boxImg
    boxImg = 0
  end if
  if len(msg) = 0 then exit sub  ' no message box to show

  dim as integer countLines = 1, maxLines = 20, fSize = FONTSIZE.X_SMALL, srch, mx, my, mb
  dim as integer msgHeight, msgWidth = 100, bWidth = 50    ' minimal width
  dim as integer maxWidth = SCREEN_X - 100
  dim as string lines(maxLines-1), tempString
  dim as FB.Image ptr lineImg
  dim as TCoord coord
  ' get message lines
  tempString = msg
  srch = instr(tempString, "|")
  do while srch
    lines(countLines-1) = left(tempString, srch-1)
    tempString = mid(tempString, srch+1)
    countLines += 1
    if countLines = maxLines then exit do
    srch = instr(tempString, "|")
  loop
  lines(countLines-1) = tempString
  ' calculate geometry
  for i as integer = 0 to countLines-1
    coord = FTsize(lines(i), font, fSize)
    if coord.x > msgWidth then msgWidth = coord.x
  next
  msgHeight = coord.y*countLines
  ' reduce box size if too big
  if msgWidth > maxWidth then msgWidth = maxWidth
  ' draw box
  boxImg = imagecreate(msgWidth+40, msgHeight+40)
  startX = (SCREEN_X-msgWidth)\2
  startY = (SCREEN_Y-msgHeight)\2
  get (startX-20, startY-20)-step(msgWidth+39, msgHeight+39), boxImg
  line (startX-20, startY-20)-step(msgWidth+39, msgHeight+39), &hffffff, bf
  line (startX-19, startY-19)-step(msgWidth+37, msgHeight+37), 0, b
  line (startX-18, startY-18)-step(msgWidth+35, msgHeight+35), 0, b
  line (startX-16, startY-16)-step(msgWidth+31, msgHeight+31), 0, b
  ' show lines
  for i as integer = 0 to countLines-1
    lineImg = printFT(lines(i), font, fSize,, stress)
    put (startX, startY+coord.y*i+2), lineImg, (0, 0)-step(msgWidth, coord.y), alpha
    imagedestroy lineImg
  next
end sub

sub sortArray(keyArray() as string, lpArray() as TListPoint, direction as integer = 1, _
         byval s as integer = -1, byval e as integer = -1)
  dim j as integer, tempString as TWString, tempLp as TListPoint
  for i as integer = lbound(keyArray)+1 to ubound(keyArray)
    tempString = keyArray(i)
    tempLp     = lpArray(i)
    j = i
    do while j > lbound(keyArray) andalso _
        (direction > 0 and keyArray(j-1) > tempString orelse _
         direction < 0 and keyArray(j-1) < tempString)
      keyArray(j) = keyArray(j - 1)
      lpArray(j)  = lpArray(j - 1)
      j = j - 1
    loop
    keyArray(j) = tempString
    lpArray(j)  = tempLp
  next
end sub

function replace(txt as TWString, srch as TWString, repl as TWString) as TWString
  ' search for 'srch' in 'txt' and replace it by 'repl'
  ' returns the new text with replacements 
  dim ret as TWString = txt, position as integer = instr(*txt.value, *srch.value)
  do while position
    ret = ret.mid(1, position-1) & repl & ret.mid(position+len(srch))
    position = instr(*ret.value, *srch.value)
  loop
  return ret
end function

function showHelpScreen(filename as string, chapter as string) as integer
  dim as FB.Image ptr tempImg = imagecreate(SCREEN_X, SCREEN_Y)
  get (0, 0)-step(SCREEN_X-1, SCREEN_Y-1), tempImg
  dim as integer mx, my, mb, ret = help(filename, chapter)
  if ret <> RETVAR.windowClosed then put (0, 0), tempImg, pset
  imagedestroy tempImg
  do
    getmouse mx, my,, mb
    sleep 1
  loop until mb < 1
  return ret
end function

function sortLetters(txt as TWString) as TWString
  dim temp as uinteger, ret as TWString = string(len(txt)+1, 0)
  for i as integer = 0 to len(txt)-1
    temp = txt[i]
    #ifdef FONT_STRESS_ESCAPE
      if temp = FONT_STRESS_ESCAPE then continue for
    #endif
    for k as integer = 0 to len(ret)-1
      if txt.value[i] < ret.value[k] then
        ret = ret.mid(1, k) & TWString(temp) & ret.mid(k+1)
        continue for, for
      end if
    next
    ret &= temp
  next
  return ret
end function
