' word flexion in German (to include into language modules)

'declare function getArticle(w as TWord, both as integer = 0) as string
'declare function number(first as integer = 0, last as integer = 0) as TWord

function conjugGerman(w as TWord, flex as integer, no as integer = 0) as string
  if w.groupname <> "verb" then return ""
  dim as TWord dw, tempW
  dim as string wordType, tempus, subj = iif(flex and INFLECT.SUBJ, " subj.", "")
  dim as integer person = (flex and 7) -1
  select case flex and (not 7) and (not INFLECT.SUBJ)
    case INFLECT.PRES      : tempus = "presente" & subj
    case INFLECT.PRETIND   : tempus = "indefinido" & subj
    case INFLECT.PRETIMP   : tempus = "imperfecto" & subj
    case INFLECT.PERF      : tempus = "perfecto comp." & subj
    case INFLECT.PQPERF    : tempus = "pluscuamperfecto" & subj
    case INFLECT.PERFANT   : tempus = "pret. anterior" & subj
    case INFLECT.FUTSIMP   : tempus = "futuro" & subj
    case INFLECT.FUTCOMP   : tempus = "futuro perf." & subj
    case INFLECT.IMPERAT   : tempus = "imperativo" & subj
    case INFLECT.IMPERATNO : tempus = "imperativo neg." & subj
    case INFLECT.COND      : tempus = "condicional" & subj
    case INFLECT.CONDPERF  : tempus = "condicional perf." & subj
    case INFLECT.PARTIC    : tempus = "participio" & subj
    case INFLECT.GERUND    : tempus = "gerundio" & subj
  end select

  ' search word in the 'german' wordlist
  foreach(TWord, tempW, globallist_add)
    if w.groupname = tempW.groupname andalso w.translate.get(no) = tempW.word then
      wordType = tempW.map.get("wordType")
      if len(wordType) = 0 orelse wordType = w.map.get("wordType" & no) then
        dw = tempW
        exit while
      end if
    end if
  nexteach
  if len(dw.word) = 0 then
    return w.translate.get(no) & " (" & str((person mod 3)+1) & ". Pers. " _
           & iif(person < 3, "sing. ", "pl. ") & tempus & ")"
  end if

  dim as string pronoun(0 to 5)    = { "ich", "du", "er/sie/es", "wir", "ihr", "sie" }
  dim as string auxPerf1(0 to 5)   = { "habe", "hast", "hat", "haben", "habt", "haben" }
  dim as string auxPerf2(0 to 5)   = { "bin", "bist", "ist", "sind", "seid", "sind" }
  dim as string auxPQPerf1(0 to 5) = { "hatte", "hattest", "hatte", "hatten", "hattet", "hatten" }
  dim as string auxPQPerf2(0 to 5) = { "war", "warst", "war", "waren", "wart", "waren" }
  dim as string auxFut(0 to 5)     = { "werde", "wirst", "wird", "werden", "werdet", "werden" }
  dim as string auxCond(0 to 5)    = { "würde", "würdest", "würde", "würden", "würdet", "würden" }
  dim as string auxPQCond1(0 to 5) = { "hätte", "hättest", "hätte", "hätten", "hättet", "hätten" }
  dim as string auxPQCond2(0 to 5) = { "wär", "wärst", "wäre", "wären", "wärt", "wären" }

  dim as string prefix = rtrim(" " & dw.map.get("prefix"))
  if tempus = "presente" or tempus = "presente subj." then
    dim as string special = "w" & (person mod 3) + 1 & iif(person > 2, "p", "s") & "pres"
    dim as string specVal = dw.map.get(special)
    if len(specVal) then return pronoun(person) & " " & specVal & prefix & iif(tempus = "presente", "", " (Subj.)")
  end if

  dim as string inf = dw.map.get("inf")
  if len(inf) = 0 then inf = dw.word
  dim as string ret, root = "", participle
  if flex = 2 or flex = 3 then root = dw.map.get("rootdiph")
  if len(root) = 0 then root = dw.map.get("root")
  if len(root) = 0 then root = left(inf, len(inf)-iif(right(inf, 2) = "en", 2, 1))
  participle = dw.map.get("participle")
  if len(participle) = 0 then
    participle = trim(prefix) & "ge" & root & iif(right(root, 1) = "t" or right(root, 1) = "d", "et", "t")
  end if
  select case flex and (not 7) and (not INFLECT.SUBJ)
    case INFLECT.PRES
      select case person
        case 0
          if right(root, 2) = "le" then
            ret = root
          else
            ret = root & "e"
          end if
        case 1
          if right(root, 1) = "s" or right(root, 1) = "z" or right(root, 2) = "ß" then
            ret = root &  "t"
          elseif right(root, 1) = "t" or right(root, 1) = "d" then
            ret = root & "est"
          else
            ret = root & "st"
          end if
        case 2, 4 : ret = root & iif(right(root, 1) = "t" or right(root, 1) = "d", "et", "t")
        case 3, 5 : ret = root & iif(right(inf, 2) = "en", "en", "n")
      end select
      ret = pronoun(person) & " " & ret & prefix
      if len(subj) then ret &= " (Subj.)"
      return ret
    case INFLECT.COND
      return pronoun(person) & " " & auxCond(person) & " " & dw.word
    case INFLECT.PRETIND, INFLECT.PRETIMP
      dim as string rootPret = dw.map.get("rootpret")
      if len(rootPret) = 0 then
        if right(root, 1) = "d" or right(root, 1) = "t" then
          rootPret = root & "ete"
        else
          rootPret = root & "te"
        end if
      else
        if person = 3 or person = 5 andalso right(rootPret, 1) <> "e" then rootPret &= "e"
      end if
      ret = pronoun(person) & " " & rootPret
      select case person
        case 1
          if right(rootPret, 1) = "d" or right(rootPret, 1) = "t" then
            ret &= "est"
          elseif right(rootPret, 1) = "s" or right(rootPret, 2) = "ß" then
            ret &= "t"
          else
            ret &= "st"
          end if
        case 4
          if right(rootPret, 1) = "d" or right(rootPret, 1) = "t" then
            ret &= "et"
          else
            ret &= "t"
          end if
        case 3, 5
          ret &= "n"
      end select
      return ret & prefix & " (" _
             & iif((flex and (not 7) and (not INFLECT.SUBJ)) = INFLECT.PRETIND, "Indefinido", "Imperfecto") _
             & iif((flex and INFLECT.SUBJ) = INFLECT.SUBJ, " Subj.", "") & ")" 
    case INFLECT.PERF
      ret = pronoun(person) & " " & auxPerf1(person) & " " & participle
      select case flex and (not 7)
        case INFLECT.PRETIND
          return ret & " (Indefinido)"
        case INFLECT.PRETIMP
          return ret & " (Imperf. Ind.)"
        case INFLECT.PRETIMP or INFLECT.SUBJ
          return ret & " (Imperf. Subj. I)"
        case INFLECT.PRETIND or INFLECT.SUBJ
          return ret & " (Imperf. Subj. II)"
        case INFLECT.PERF
          return ret & " (Perfecto)"
      end select
    case INFLECT.CONDPERF
      return pronoun(person) & " " & auxPQCond1(person) & " " & participle
    case INFLECT.PQPERF
      return pronoun(person) & " " & auxPQPerf1(person) & " " & participle
    case INFLECT.PERFANT
      return pronoun(person) & " " & auxPQPerf1(person) & " " & participle & " (Perf. anterior)"
    case INFLECT.FUTSIMP
      return pronoun(person) & " " & auxFut(person) & " " & dw.word & iif(len(subj), " (Subj.)", "")
    case INFLECT.FUTCOMP
      return pronoun(person) & " " & auxFut(person) & " " & participle & " haben" & iif(len(subj), " (Subj.)", "")
    case INFLECT.IMPERAT, INFLECT.IMPERATNO
      dim as string no = iif(flex and (not 7) = INFLECT.IMPERATNO, " nicht", "")
      select case person
        case 1
          dim as string imperativ = dw.map.get("impsg")
          if len(imperativ) = 0 then imperativ = root & "e"
          return firstUpper(imperativ) & no & prefix & "!"
        case 2, 5
          return firstUpper(inf) & " Sie" & no & prefix & "! (" & iif(person = 2, "sg.)", "pl.)")
        case 3
          return "Lasst uns" & no & " " & dw.word & "!"
        case 4
          dim as string imperativ = dw.map.get("imppl")
          if len(imperativ) = 0 then imperativ = root & "t"
          return firstUpper(imperativ) & no & prefix & "!"
      end select
    case INFLECT.PARTIC
      return participle
    case INFLECT.GERUND
      return dw.word & "d"
  end select
  return "-"
end function

function declinGerman(w as TWord, flex as integer, no as integer = 0) as string
  if w.groupname = "adjective" then
    dim as string ret = w.translate.get(no)
    select case flex
      case INFLECT.MASC or INFLECT.SING
        return ret
      case INFLECT.FEM  or INFLECT.SING
        return ret & " (fem. sing.)"
      case INFLECT.MASC or INFLECT.PL
        return ret & " (mask. pl.)"
      case INFLECT.FEM  or INFLECT.PL
        return ret & " (fem. pl.)"
    end select
  elseif w.groupname = "noun" then
    dim as TWord dw, tempW
    dim as string wordType, num = w.map.get("num"), genus = w.map.get("genus"), fem = w.map.get("f")
  
    ' search word in the 'german' wordlist
    foreach(TWord, tempW, globallist_add)
      if w.groupname = tempW.groupname andalso w.translate.get(no) = tempW.word then
        wordType = tempW.map.get("wordtype")
        if len(wordType) = 0 orelse wordType = w.map.get("wordtype" & no) then
          dw = tempW
          exit while
        end if
      end if
    nexteach
    if len(dw.word) = 0 then
      return w.translate.get(no) & " (" & iif((flex and INFLECT.FEM) = INFLECT.FEM, "fem.", "mask.") _
             & iif((flex and INFLECT.PL) = INFLECT.PL, "pl.)", "sing.)")
    end if
    if (flex and INFLECT.NUM) = INFLECT.PL then
      dim as string pl = dw.map.get("pl")
      if len(pl) then return pl else return dw.word & " (pl.)"
    else
      if dw.map.get("num") = "pl" then return dw.word & " (sg.)" else return dw.word
    end if
  end if
  return ""
end function
