' basic module for language training Italian-German

const SCREEN_X = 800, SCREEN_Y = 640, B_COUNT = 6
const moduleVersionMajor = 2, moduleVersionMinor = 4, moduleVersionPatch = 180112
const LANGTITLE1 = "Italienisch",   LANGTITLE2 = "Deutsch"
const LANGSHORT1 = "it",            LANGSHORT2 = "de"

dim shared as string LANGUAGE, MODULE_PATH, WORDS_VERSION
LANGUAGE = mid(__PATH__, instrrev(__PATH__, any "/\")+1)
MODULE_PATH = exepath & "/modules/" & LANGUAGE & "/"

' TODO: various corrections (see below)

' options to modify text to a different form
' BE SURE NOT TO USE CONFLICTING OPTIONS!
const MODIFY_TOLOWER       =   1, _ ' change to lower case if possible
      MODIFY_TOUPPER       =   2, _ ' change to upper case if possible
      MODIFY_NOPARANTHESES =   4, _ ' omit parantheses ( and ) 
      MODIFY_NOSPECIAL     =   8, _ ' omit everything exept a-z, A-Z, 0-9 and language related keys (e. g. umlauts in German)
      MODIFY_NOACUTE       =  16, _ ' change é to e and so on (umlauts remain unaffected)
      MODIFY_NOUMLAUT      =  32, _ ' change ß to ss (or SS if MODIFY_TOUPPER), Ä to Ae (or AE or ae) and so on
      MODIFY_NONUMBERS     =  64, _ ' omit 0-9
      MODIFY_SORTABLE      = 128, _ ' various changes like lower case, ä->ae, é->e, ñ->n| (to be sorted between n and o) ...
      MODIFY_ALPHANUM      = 256    ' everthing exept a-z, A-Z and 0-9 is changed to an underline character

namespace INFLECT
  const PERS1 = 1, PERS2 = 2, PERS3 = 3
  const SINGULAR = 0, PLURAL = 3
  const PRES = 0
  const MASK = 0, FEM = 1, SING = 0, PL = 2, ADVERB = 4
  const FULL = 1, MINIMUM = 2
end namespace

const COLOR_GRAY = rgb(127, 127, 127)

#include once "include/fontExt.bi"
dim shared font as FT_FaceFam

#include once "include/globalFunctions.bi"
dim shared as TExercise globalExercArray(), globalOptArray()

declare function conjugItalian(verb as TWord, inflection as integer) as TWord
declare function conjugGerman(verb as TWord, inflection as integer, no as integer = 0) as string
declare function declinItalian(noun as TWord, inflection as integer) as string
declare function declinGerman(noun as TWord, inflection as integer, no as integer = 0) as string
declare function flexionCount(w as TWord, opt as integer = INFLECT.FULL) as integer
declare function getArticle(w as TWord, both as integer = 0) as string
declare sub init(mainModule as TModule ptr)
declare function modifyTxt(txt as string, opt as integer = MODIFY_TOLOWER or MODIFY_NOPARANTHESES) as string
declare function number(first as integer = 0, last as integer = 0) as TWord
declare sub speak(txt as string, no as integer = 1, byref ready as integer = 0)
declare sub speakThread(txt as string, no as integer = 1)
declare sub wordInfo(byval w as TWord)

#include once "specific.bi"
#include once "include/common/lang-basic.bi"

#include once "include/common/editWords.bi"
#include once "include/common/learning.bi"
#include once "include/common/wordtest.bi"
#include once "include/common/diagram.bi"
#include once "include/common/dictionary.bi"
#include once "incgui.bi"

'#include once "export.bi"

' games
#include once "include/common/pairs.bi"
#include once "include/common/crossword.bi"
#include once "include/common/hangman.bi"
globalExercises.add(TExercise("games", localize("gamesTitle"), "full", MENUTYPE.exercTwoRowsExt, _
                    @createCategory, 0, 0, localize("gamesHeadline")))

function conjugItalian(verb as TWord, inflection as integer) as TWord
  ' TODO - ONLY INFLECT.PRES IS SUPPORTED
  dim as TWord ret
  if len(verb.word) < 3 then return ret    ' error
  dim as string retWord, retTrans, tempString, root, regularEnding
  dim as integer tempint = (inflection and 7)-1, mark = 1
  dim as string infl = ((tempint mod INFLECT.PLURAL) + 1) _
                       & iif(tempint\INFLECT.PLURAL, "p", "s"), reflex = ""
  if verb.word.mid(len(verb.word) - 1) = "si" then
    select case as const inflection and 7
      case INFLECT.PERS1 + INFLECT.SINGULAR : reflex = "mi "
      case INFLECT.PERS2 + INFLECT.SINGULAR : reflex = "ti "
      case INFLECT.PERS3 + INFLECT.SINGULAR : reflex = "si "
      case INFLECT.PERS1 + INFLECT.PLURAL   : reflex = "ci "
      case INFLECT.PERS2 + INFLECT.PLURAL   : reflex = "vi "
      case INFLECT.PERS3 + INFLECT.PLURAL   : reflex = "si "
    end select
  end if

  ' looking for irregular form
  select case as const inflection and (not 7)
    case INFLECT.PRES    : infl &= "pres"    : root = verb.map.get("rootpres")
  end select
  retWord  = verb.map.get("w" & infl)
  retTrans = verb.map.get("t" & infl)
  if retWord = "" then
    if len(root) = 0 then root = verb.map.get("root") : mark = 0
    ' regular form
    select case inflection and (not 7)
      ' PRESENTE DE INDICATIVO
      case INFLECT.PRES
        if (inflection and 7) <> INFLECT.PERS1+INFLECT.PLURAL and _
           (inflection and 7) <> INFLECT.PERS2+INFLECT.PLURAL then
          tempString = verb.map.get("rootdiph")
          if len(tempString) then root = tempString : mark = -1
        end if
        if len(root) = 0 then root = verb.word.mid(1, len(verb.word)-2+(len(reflex)>0))
        tempInt = len(root)-1
        select case as const inflection and 7
          case INFLECT.PERS1 + INFLECT.SINGULAR
            regularEnding = "o"
          case INFLECT.PERS2 + INFLECT.SINGULAR
            regularEnding = "i"
          case INFLECT.PERS3 + INFLECT.SINGULAR
            if root[tempInt] = 97 then regularEnding = "a" else regularEnding = "e"
          case INFLECT.PERS1 + INFLECT.PLURAL
            regularEnding = "iamo"
          case INFLECT.PERS2 + INFLECT.PLURAL
            if root[tempInt] = 97 then
              regularEnding = "ate"
            elseif root[tempInt] = 101 then
              regularEnding = "ete"
            else
              regularEnding = "ite"
            end if
          case INFLECT.PERS3 + INFLECT.PLURAL
            if root[tempInt] = 97 then regularEnding = "ano" else regularEnding = "ono"
        end select
    end select
    retWord = left(root, len(root)-1) & regularEnding
  end if
  retWord = reflex & retWord
  if retTrans = "" then retTrans = "<" & verb.translate.get(0) & ">"
  ret = TWord(retWord, retTrans, "verb")
  ret.mark = mark
  return ret
end function

function conjugGerman(verb as TWord, inflection as integer, no as integer = 0) as string
  dim as string ret, transl = verb.translate.get(no), root
  dim as integer tempint = (inflection and 7)-1
  if len(transl) < 3 then return ""    ' error
  dim as string infl = ((tempint mod INFLECT.PLURAL) + 1) _
                       & iif(tempint\INFLECT.PLURAL, "p", "s")
  dim as string nt = iif(no = 0, "t", "t" & no)
  dim as string pronoun(0 to 5)    = { "ich", "du", "er/sie/es", "wir", "ihr", "sie" }
  dim as string suffix(0 to 5)     = { "e", "st", "t", "en", "t", "en" }
  dim as string sufPret(0 to 5)    = { "", "st", "", "n", "t", "n" }
  dim as string auxPerf1(0 to 5)   = { "habe", "hast", "hat", "haben", "habt", "haben" }
  dim as string auxPerf2(0 to 5)   = { "bin", "bist", "ist", "sind", "seid", "sind" }
  dim as string auxPQPerf1(0 to 5) = { "hatte", "hattest", "hatte", "hatten", "hattet", "hatten" }
  dim as string auxPQPerf2(0 to 5) = { "war", "warst", "war", "waren", "wart", "waren" }
  dim as string auxFut(0 to 5)     = { "werde", "wirst", "wird", "werden", "werdet", "werden" }
  dim as integer person = (inflection-1) mod 6

  ' looking for irregular form
  select case as const inflection and (not 7)
    case INFLECT.PRES                 : infl &= "pres"    : root = verb.map.get(nt & "rootpres")
  end select
  ret = verb.map.get("t" & infl)
'if len(ret) = 0 then ret = verb.map.get("root") : mark = 0
  ' regular form
  select case inflection and (not 7)
    ' PRÄSENS
    case INFLECT.PRES
      if len(ret) = 0 then ret = left(transl, len(transl)-2)
      ret = pronoun(person) & " " & ret & suffix(person)
    ' PRÄTERITUM
  end select
  return ret
end function

function declinItalian(noun as TWord, inflection as integer) as string
  ' not implemented yet
  return noun.word
end function

function declinGerman(noun as TWord, inflection as integer, no as integer = 0) as string
  ' not implemented yet
  return noun.translate.get(0)
end function

function flexionCount(w as TWord, opt as integer = INFLECT.FULL) as integer
  select case w.groupname
    case "verb"
      select case opt
        case INFLECT.MINIMUM : return 6
        case else            : return 6
      end select
    case "noun"      : return 3
    case "adjective" : return 4
    case else        : return -1
  end select
end function

function getArticle(w as TWord, both as integer = 0) as string
  ' returns the article(s) of a noun, followed by a space character
  ' If there is no article (because w is not a noun), the function returns an empty string
  ' both <> 0: if there exist (different) forms for m and f, return il/la, or i/le respectively
  enum first
    none = 0, vowel, consonant, special
  end enum
  if w.groupname <> "noun" then return ""
  dim as string genus = w.map.get("genus"), article, num
  dim as first firstLetter
  select case lcase(w.word.mid(1, 1))
    ' "ordinary" consonant
    case "b", "c", "d", "f", "h", "j", "k", "l", "m", "n", "q", "r", "t", "v", "w"
      firstLetter = first.consonant
    ' special chars
    case "x", "y", "z"
      firstLetter = first.special
    ' special combinations
    case "g"
      if lcase(mid(w.word, 2, 1)) = "n" then firstLetter = first.special else firstLetter = first.consonant
    case "p"
      if lcase(mid(w.word, 2, 1)) = "s" then firstLetter = first.special else firstLetter = first.consonant
    case "s"
      if instr(lcase(mid(w.word, 2, 1)), "bcdfghjklmnpqrstvwxyz") then 
        firstLetter = first.special
      else
        firstLetter = first.consonant
      end if
    ' vowels
    case else
      firstLetter = first.vowel
  end select
  num = w.map.get("num")
  select case genus
    case "m"
      select case firstLetter
        case first.consonant
          article = iif(num = "pl", "i", "il")
          if both andalso len(w.map.get("f")) then article &= iif(num = "pl", "/le", "/la")
        case first.vowel
          article = iif(num = "pl", "gli", "l'")
          if both andalso num = "pl" andalso len(w.map.get("f")) then article &= "/le"
        case first.special
          article = iif(num = "pl", "gli", "lo")
          if both andalso len(w.map.get("f")) then article &= iif(num = "pl", "/le", "/la")
      end select
    case "f"
      select case firstLetter
        case first.vowel
          article = iif(num = "pl", "le", "l'")
        case else
          article = iif(num = "pl", "le", "la")
      end select
    case else
      select case firstLetter
        case first.consonant
          article = iif(num = "pl", "i/le", "il/la")
        case first.vowel
          article = iif(num = "pl", "gli/le", "l'")
        case first.special
          article = iif(num = "pl", "gli/le", "lo/la")
      end select
  end select
  return article & " "
end function

sub init(mainModule as TModule ptr) export
  initModule(mainModule)
  dim as TList fullList
  fullList.listname = "complete"
  fullList.add , 0, &hffffff
  fullList.save
  fullList.listname = "full"
  fullList.del 0, &hffffff, "group", "cardinal"
  fullList.save
  initDefaultLists(fullList)
  mainWindow
  exitModule
end sub

function infoDefault(exercise as TExercise) as string
  ' default information string for exercises (shown in the main window)
  dim as TList ptr listPtr = iif(exercise.currentList.size = 0, @exercise.fullList, @exercise.currentList)
  return "ausstehend: " & listPtr->notChecked & " von " & listPtr->size & " Wörtern"
end function

function infoEmpty(exercise as TExercise) as string
  ' empty information string for exercises (for the main window)
  return ""
end function

function modifyTxt(txt as string, opt as integer = MODIFY_TOLOWER or MODIFY_NOPARANTHESES) as string
  ' modify input (see modify options at the top of the file)
  ' Using conflicting options may produce unexpected results
  dim as string ret
  dim as integer i = 0
  do until i >= len(txt)
    select case as const txt[i]
      case 0 to 31                                    ' ignore control chars
      case 65 to 90                                   ' upper case
        ret &= chr(txt[i] + iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), 32, 0))
      case 97 to 122                                  ' lower case
        ret &= chr(txt[i] - iif(opt and MODIFY_TOUPPER, 32, 0))
      case 40, 41                                     ' parentheses
        if opt and MODIFY_ALPHANUM then
          ret &= "_"
        else
          if opt and (MODIFY_NOPARANTHESES or MODIFY_NOSPECIAL) = 0 then ret &= chr(txt[i])
        end if
      case 48 to 57                                   ' numbers
        if (opt and MODIFY_NONUMBERS) = 0 then ret &= chr(txt[i])
      case 32 to 39, 42 to 47, 58 to 64, 91 to 96, 123 to 127   ' special keys
        if opt and MODIFY_ALPHANUM then
          ret &= "_"
        else
          if (opt and MODIFY_NOSPECIAL) = 0 then ret &= chr(txt[i])
        end if
      case 195
        i += 1
        if i >= len(txt) then exit do
        if opt and MODIFY_ALPHANUM then ret &= "_" : exit select
        select case as const txt[i]
          case 132 ' Ä
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              if opt and (MODIFY_SORTABLE or MODIFY_TOLOWER) then
                ret &= "ae"
              elseif opt and MODIFY_TOUPPER then
                ret &= "AE"
              else
                ret &= "Ae"
              end if
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ä", "Ä")
            end if
          case 164 ' ä
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              ret &= iif(opt and MODIFY_TOUPPER, "AE", "ae")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ä", "ä")
            end if
          case 150 ' Ö
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              if opt and (MODIFY_SORTABLE or MODIFY_TOLOWER) then
                ret &= "oe"
              elseif opt and MODIFY_TOUPPER then
                ret &= "OE"
              else
                ret &= "Oe"
              end if
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ö", "Ö")
            end if
          case 182 ' ö
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              ret &= iif(opt and MODIFY_TOUPPER, "OE", "oe")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ö", "ö")
            end if
          case 156 ' Ü
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              if opt and (MODIFY_SORTABLE or MODIFY_TOLOWER) then
                ret &= "ue"
              elseif opt and MODIFY_TOUPPER then
                ret &= "UE"
              else
                ret &= "Ue"
              end if
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ü", "Ü")
            end if
          case 188 ' ü
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              ret &= iif(opt and MODIFY_TOUPPER, "UE", "ue")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ü", "ü")
            end if
          case 159 ' ß
            if opt and (MODIFY_SORTABLE or MODIFY_NOUMLAUT) then
              ret &= iif(opt and MODIFY_TOUPPER, "SS", "ss")
            else
              ret &= "ß"
            end if
          case 129 ' Á
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), "a", "A")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "á", "Á")
            end if
          case 161 ' á
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and MODIFY_TOUPPER, "A", "a")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Á", "á")
            end if
          case 137 ' É
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), "e", "E")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "é", "É")
            end if
          case 169 ' é
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and MODIFY_TOUPPER, "E", "e")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "É", "é")
            end if
          case 141 ' Í
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), "i", "I")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "í", "Í")
            end if
          case 173 ' í
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and MODIFY_TOUPPER, "I", "i")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "´I", "´i")
            end if
          case 147 ' Ó
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), "o", "O")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ó", "Ó")
            end if
          case 179 ' ó
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and MODIFY_TOUPPER, "O", "o")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ó", "ó")
            end if
          case 154 ' Ú
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), "u", "U")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ú", "Ú")
            end if
          case 186 ' ú
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and MODIFY_TOUPPER, "U", "u")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ú", "ú")
            end if
          case 145 ' Ñ
            if opt and MODIFY_SORTABLE then
              ret &= iif(opt and MODIFY_TOLOWER, "n|", "N|")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ñ", "Ñ")
            end if
          case 177 ' ñ
            if opt and MODIFY_SORTABLE then
              ret &= iif(opt and MODIFY_TOUPPER, "N|", "n|")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ñ", "ñ")
            end if
          case else    ' in use: (192,161): ¡   (192,191): ¿
            if (opt and (MODIFY_SORTABLE or MODIFY_NOSPECIAL)) = 0 then ret &= "¡"
        end select
    end select
    i += 1
  loop
  return ret
end function

sub speak(txt as string, no as integer = 1, byref ready as integer = 0)
  ' TODO (or not ...)

  'mutexlock speakLock
  'dim search as integer, t as string = txt
  'ready = -1
  'mutexunlock speakLock
  'search = instr(t, "ll")
  'do while search
  '  t = left(t, search-1) & "y" & mid(t, search+2)
  '  search = instr(t, "ll")
  'loop
  'search = instr(t, "rr")
  'do while search
  '  t = left(t, search) & "'" & mid(t, search+1)
  '  search = instr(t, "rr")
  'loop
  'shell "espeak -vmb-es" & no & " """ & t & """ -s100"
end sub

sub speakThread(txt as string, no as integer = 1)
' see speak()
'  dim as integer ready = 0, r
'  dim as any ptr thread = threadcall speak(txt, no, ready)
'  do
'    mutexlock speakLock
'    r = ready
'    mutexunlock speakLock
'  loop until r
'  threaddetach(thread)
end sub

sub wordInfo(byval w as TWord)
  ' TODO - ONLY INFLECT.PRES IS IMPLEMENTED
  dim as TWord tempWord
  if w.groupname = "verb" then
    dim as integer conjX = 50, conjPresX = 50, conjPresY = 100', conjPretX = 300, conjPretY = 100
    'dim as integer conjImperfX = 550, conjImperfY = 100
    dim as integer conjHeadY = 50, conjLines = 20, conjSize = FONTSIZE.X_SMALL
    dim as string person(1 to 6) = { "io", "tu", "lui/lei/Lei", "noi", "voi", "loro" }
    screenlock
    cls
    dim as string tempString = w.translate.get(0)
    for i as integer = 1 to w.translate.size - 1
      tempString &= ", " & w.translate.get(i)
    next
    printFT conjX, conjHeadY, rtrim(w.word & " " & w.map.get("obj")) & "  -  " & tempString, font, conjSize
    printFT conjPresX, conjPresY, "presente de indicativo", font, conjSize
    for i as integer = 1 to 3
      tempWord = conjugItalian(w, INFLECT.PRES or i)
      printFT conjPresX, conjPresY+(i+.5)*conjLines, person(i), font, conjSize
      printFT conjPresX+80, conjPresY+(i+.5)*conjLines, tempWord.word, _
              font, conjSize, iif(tempWord.mark, &hff0000, 0)
      tempWord = conjugItalian(w, INFLECT.PRES or (i+INFLECT.PLURAL))
      printFT conjPresX, conjPresY+(i+3.5)*conjLines, person(i+3), font, conjSize
      printFT conjPresX+80, conjPresY+(i+3.5)*conjLines, tempWord.word, _
              font, conjSize, iif(tempWord.mark, &hff0000, 0)
    next
    screenunlock
    getkey
  end if
end sub
